﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=8.5
@EndOfDesignText@
'NOTE: This class is based on double precision numbers.  The statistical routines and matrix inversion require double precision.
'Use display(0) for integer display of a dMatrix, display(5) to show 5 decimal places, etc.
'For 3 dimensions, dMatrix can be used as items on a list of matrices, or for 4 dimensions as items on a a list of lists of matrices, etc.

'For scalars, use a 1x1 dMatrix
'For vectors use either a row vector (1xn dMatrix) or a column vector (nx1 dMatrix) depending on the situation

'Some of the row operations do not make sense in statistical operations 
'		(for example the mean of rows where the value in the row are differently scaled variables)
'		(it might make sense if the columns are the same measure repeated over time)

'The instances of dMatrix will have an Id when they are created, this Id propagates to derived matrices and is used for error tracing

'There are 60 public methods - iterators are used.  Row operations are done on the transpose of the columns and transposed back.
'The public "apply" iterator is a powerful way to extend the functions that can be performed the items of a dMatrix

Sub Class_Globals
	Type sortPairDouble(index As Int, value As Double)
	Public id As String
	Public nr, nc As Int
	Public x(,) As Double
	Public firstMethodApplied As String
	Public lastMethodApplied As String

	Private mCallBack As Object			'for error reporting
	Private byRowCol As String = "column"
	Private alertCount As Int = 0
	
	Private functionAKA As Map
End Sub

'Global: Private dMx as dMatrix  dMx.Initialize(Me, "")  provides accesses to the various contructors and utilities
Public Sub Initialize(callBack As Object, name As String) As dMatrix
	mCallBack = callBack
	id = name
	nr = 0: nc = 0
	Dim x(0, 0) As Double
	functionAKA = CreateMap("addAmt":"plus", "minusAmt":"minus", "multAmt":"times", "divideAmt":"divide", "clone":"copy")
	Return Me
End Sub

Private Sub methods(M As dMatrix, method As String) As dMatrix
	If firstMethodApplied = "" Then 
		M.firstMethodApplied = "." & method
		M.lastMethodApplied = ""
	Else 
		M.firstMethodApplied = firstMethodApplied
		M.lastMethodApplied = "…" & method
	End If
	Return M
End Sub

'"id" property of dMatrix is used for display and tracing errors, rename resets old traces
Public Sub rename(name As String) As dMatrix
	id = name
	firstMethodApplied = ""
	lastMethodApplied = ""
	Return Me
End Sub

'dNew method is the most generic constructor of dMatrix, name becomes "id"
Public Sub dNew(name As String, m As Int, n As Int, v() As Double) As dMatrix
	Dim newM As dMatrix: newM.Initialize(mCallBack, name)
	newM.nr = m: newM.nc = n
	Dim newX(m, n) As Double
	Dim index As Int = -1
	For i = 0 To m -1
		For j = 0 To n - 1
			index = index + 1
			If index = v.length  Then index = 0
			newX(i, j) = v(index)
		Next
	Next
	newM.x = newX
	Return newM
End Sub

'dValue method is a common constructor of dMatrix, usually to prepare a Amount filled matrix of the right shape
Public Sub dValue(name As String, m As Int, n As Int, value As Double) As dMatrix
	Return dNew(name, m, n, Array As Double(value))
End Sub

'dZero method is a common constructor of dMatrix, usually to prepare a zero filled matrix of the right shape
Public Sub dZero(name As String, m As Int, n As Int) As dMatrix
	Return dNew(name, m, n, Array As Double(0))
End Sub

'dScalar constructs a 1x1 matrix with "z" as the entry
Public Sub dScalar(name As String, z As Double) As dMatrix
	Return dNew(name, 1, 1, Array As Double(z))
End Sub

'dDiagonal constructs a diagonal matrix with the items in the double array v() cyclically filling the diagonal
Public Sub dDiagonal(name As String, m As Int, v() As Double) As dMatrix
	Dim newM As dMatrix: newM.Initialize(mCallBack, name)
	newM.nr = m: newM.nc = m
	Dim newX(m, m) As Double
	Dim index As Int = -1
	For i = 0 To m -1
		index = index + 1
		If index = v.length  Then index = 0
		newX(i, i) = v(index)
	Next
	newM.x = newX
	Return newM
End Sub

'dIdentity constructs a diagonal matrix with the ones on the diagonal
Public Sub dIdentity(name As String, m As Int) As dMatrix
	Return dDiagonal(name, m, Array As Double(1))
End Sub

'dString takes a smart string and parses it into a matrix that is wrapped in a dMatrix object
Public Sub dString(name As String, source As String) As dMatrix
	Dim newM As dMatrix: newM.Initialize(mCallBack, name)
	Dim v() As String = Regex.Split(CRLF, source)
	Dim m, n As Int
	'find the max number of items per line
	For Each q As String In v
		q = q.replace(TAB, "")
		If q.trim <> "" Then 
			m = m + 1
			Dim w() As String = Regex.Split("\,", q)
			If w.Length > n Then n = w.length
		End If
	Next
	'define and populate the matrix
	newM.nr = m: newM.nc = n
	Dim newX(m, n) As Double
	Dim  index As Int = -1
	For Each q As String In v
		q = q.replace(TAB, "")
		If q.trim <> "" Then 
			index = index + 1
			Dim w() As String = Regex.Split("\,", q)
			For j = 0 To Min(w.length -1, n - 1)
				Dim s As String = w(j)
				If IsNumber(s) = False Then s = "NaN"
				newX(index, j) = s
			Next
			For j = Min(w.length, n) To n - 1		'fill empty value with zero
				newX(index, j) = 0
			Next
		End If
	Next
	newM.x = newX
	Return newM
End Sub

'dRandom constructs a dMatrix object with entries that are random numbers between lower and upper with nDecimal places
Public Sub dRandom(name As String, m As Int, n As Int, lower As Double, upper As Double, nDecimal As Int) As dMatrix
	Return dNew(name, m, n, randomVector(m * n, lower, upper, nDecimal))
End Sub

Private Sub randomVector(n As Int, lower As Double, upper As Double, nDecimal As Int) As Double()
	Dim multiplier As Double = Power(10, nDecimal)
	Dim v(n) As Double
	For i = 0 To n - 1
		v(i) = Rnd(lower * multiplier, upper * multiplier) / multiplier
	Next
	Return v
End Sub

'dWrap takes an ordinary matrix and constructs a dMatrix object
Public Sub dWrap(name As String, mat(,) As Double) As dMatrix
	Dim z() As Double = asList(mat).Get(0)
	Dim m As Int = mat.length
    Dim n As Int = z.Length
	Dim newM As dMatrix: newM.Initialize(mCallBack, name)
	newM.nr = m: newM.nc = n
	Dim newX(m, n) As Double
	For i = 0 To m - 1
		For j = 0 To n - 1
			newX(i, j) = mat(i, j)
		Next
	Next
	newM.x = newX
	Return newM
End Sub

'dList takes a list of equal length vectorss and constructs a dMatrix object
Public Sub dList(name As String, aList As List) As dMatrix
	Dim z() As Double = aList.Get(0)
	Dim m As Int = aList.Size
    Dim n As Int = z.Length
	Dim newM As dMatrix: newM.Initialize(mCallBack, name)
	newM.nr = m: newM.nc = n
	Dim newX(m, n) As Double
	For i = 0 To m - 1
		z = aList.Get(i)
		For j = 0 To n - 1
			If z.Length - 1 < j Then 
				newX(i, j) = 0
			Else
				newX(i, j) = z(j)
			End If
		Next
	Next
	newM.x = newX
	Return newM
End Sub

'For flexibility (Pages or no Pages), the output is handled by "myLog" which has to be in the calling program
Public Sub display(nDecimal As Int) As dMatrix
	Dim eps As Double = 1 / Power(10, nDecimal + 2)		'used to avoid the minus sign in -0.0000
	Dim sb As StringBuilder: sb.initialize
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			Dim val As Double = x(i, j)
			If Abs(val) < eps Then val = 0
			Dim s As String = NumberFormat2(val, 1, nDecimal, nDecimal, False)
			If s = "NaN" Then s = Chr(0x25C8)
			sb.Append(s).Append(TAB)
		Next
		sb.Remove(sb.Length -1, sb.Length).append(CRLF)
	Next
	CallSub3(mCallBack, "myLog", sb.toString, $"${id}${firstMethodApplied}${lastMethodApplied} [${nr}, ${nc}]"$)
	Return Me
End Sub

'For flexibility (Pages or no Pages), the error handling sub "matrixErrors" should be in the calling program
Private Sub errorMessage(s As String) As dMatrix
	CallSub2(mCallBack, "matrixErrors", s) 
	Return Me
End Sub

Private Sub isEmpty As Boolean
	If nr = 0 Or nc = 0 Then 
		errorMessage($"?${id} is empty."$)
		Return True
	Else
		Return False
	End If
End Sub

'getDiagonal selects the diagonal and constructs a min(m,n) x 1 dmatrix object
Public Sub getDiagonal As dMatrix
	If isEmpty Then Return Me
	Dim m As Int = Min(nr, nc)
	Dim newM As dMatrix = dValue(id, m, 1, 0)
	For i = 0 To m - 1
		newM.x(i, 0) = x(i, i)
	Next
	Return methods(newM, "diagonal")
End Sub

'getColumn selects the intended column and constructs a 1xn dmatrix object
Public Sub getColumn(index As Int) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, nr, 1, 0)
	For i = 0 To nr - 1
		newM.x(i, 0) = x(i, index)
	Next
	Return methods(newM, "column" & index)
End Sub

'getRow selects the intended row and constructs a mx1 dmatrix object
Public Sub getRow(index As Int) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, 1, nc, 0)
	For j = 0 To nc - 1
		newM.x(0, j) = x(index, j)
	Next
	Return methods(newM, "row" & index)
End Sub

'selects elements based on each pair in two arrays of indices
Public Sub getSubset(rIndices() As Int, cIndices() As Int) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, rIndices.length, cIndices.length, 0)
	For i = 0 To newM.nr - 1
		For j = 0 To newM.nc - 1
			newM.x(i, j) = x(rIndices(i), cIndices(j))
		Next
	Next
	Return methods(newM, "subset")
End Sub

'Transforms construct a dMatrix object with row and columns transposed newX(i, j) = oldX(j, i)
Public Sub transpose As dMatrix
	If isEmpty Then Return Me
	Dim M As dMatrix = dValue(id, nc, nr, 0)
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			M.x(j, i) = x(i, j)
		Next
	Next
	Return methods(M, "transpose")
End Sub

'Basic scalar operations allow for various combinations of matrix sizes - but they need to be compatible/commensurate
Private Sub getOpType(M As dMatrix, op As String) As String
	Dim optyp As String = ""
	If M.nc = 1 And M.nr = 1 Then
		optyp = "scalar"
	Else If M.nr = 1 And M.nc = nc Then
		optyp = "row"
	else If M.nc = 1 And M.nr = nr Then
		optyp = "column"
	Else If M.nc <> nc Or M.nr <> nr Then 
		errorMessage($"${op}: One matrix ${id} has dimensions (${nr},${nc}), but the second matrix ${M.id} has (${M.nr},${M.nc})."$)
	Else
		optyp = "cell"
	End If
	Return optyp
End Sub

'a function (in "context") is applied to each item in the dMatrix object, it uses the parameter object array
Public Sub iterate(context As Object, function As String, parameters() As Object) As dMatrix	'ignore
	If (nr = 0 Or nc = 0) And function <> "clone"  Then Return errorMessage($"?${id} is empty."$)
	Dim fname As String = function
	Dim operand As String = ""
	If functionAKA.ContainsKey(function) Then fname = functionAKA.Get(function)
	Dim newM As dMatrix = dValue(id, nr, nc, 0)
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			If parameters.length = 2 And GetType(parameters(1)).Contains("String") Then
				Dim M As dMatrix = parameters(0)
				operand = M.id
				Dim typ As String = parameters(1)
				Select typ
					Case "scalar": newM.x(i, j) = CallSub3(context, function, x(i, j), Array(M.x(0, 0)))
					Case "row":	newM.x(i, j) = CallSub3(context, function, x(i, j), Array(M.x(0, j)))
					Case "column": newM.x(i, j) = CallSub3(context, function, x(i, j), Array(M.x(i, 0)))
					Case "cell": newM.x(i, j) = CallSub3(context, function, x(i, j), Array(M.x(i, j)))
				End Select
			Else	
				newM.x(i, j) = CallSub3(context, function, x(i, j), parameters)
			End If
		Next
	Next
	Return methods(newM, fname & "." & operand)
End Sub

'copies all aspects of the dMatrix object to a new dMatrix object
Public Sub copy As dMatrix
	Return iterate(Me, "clone", Array())
End Sub

'adds the elements of a dMatrix object (M) and returns a new dMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub plus(M As dMatrix) As dMatrix
	Dim opTyp As String = getOpType(M, "plus")
	If opTyp = "" Then Return Me Else Return iterate(Me, "addAmt", Array(M, opTyp))
End Sub

'muliplies elements by a elements of dMatrix object (M) and returns a new dMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub times(M As dMatrix) As dMatrix
	Dim opTyp As String = getOpType(M, "times")
	If opTyp = "" Then Return Me Else Return iterate(Me, "multAmt", Array(M, opTyp))
End Sub

'subtracts elements by elements of a dMatrix object (M) and returns a new dMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub minus(M As dMatrix) As dMatrix
	Dim opTyp As String = getOpType(M, "minus")
	If opTyp = "" Then Return Me Else Return iterate(Me, "minusAmt", Array(M, opTyp))
End Sub

'divides elements by elements of a dMatrix object (M) and returns a new dMatrix object
'M can must have commensurate dimensions (possibly single row, single column, or 1x1)
Public Sub divide(M As dMatrix) As dMatrix
	Dim opTyp As String = getOpType(M, "divide")
	If opTyp = "" Then Return Me Else Return iterate(Me, "divideAmt", Array(M, opTyp))
End Sub

'takes the absolute value of elements and returns a dMatrix object with the results
Public Sub absValue As dMatrix
	Return iterate(Me, "absolute", Array())
End Sub

'negates elements and returns a dMatrix object with the results
Public Sub negate As dMatrix
	Return iterate(Me, "negative", Array())
End Sub

'squares elements and returns a dMatrix object with the results
Public Sub squared As dMatrix
	Return iterate(Me, "square", Array())
End Sub

'takes the square root of elements and returns a dMatrix object with the results
Public Sub sqRoot As dMatrix
	Return iterate(Me, "squareRoot", Array())
End Sub

'adds a amount to each element and returns a dMatrix object with the results
Public Sub plusAmount(a As Double) As dMatrix
	Return iterate(Me, "addAmt", Array(a))
End Sub

'multiplies each by an amount and returns a dMatrix object with the results
Public Sub timesAmount(a As Double) As dMatrix
	Return iterate(Me, "multAmt", Array(a))
End Sub

'similar to iterate but callBack context is set calling context
Public Sub apply(function As String, parameters() As Object) As dMatrix			'ignore
	Return iterate(mCallBack, function, parameters)
End Sub

Private Sub clone(z As Double, pars() As Object) As Double	'ignore
	Return z
End Sub

Private Sub addAmt(z As Double, pars() As Object) As Double	'ignore
	Dim amt As Double = pars(0)
	Return z + amt
End Sub

Private Sub multAmt(z As Double, pars() As Object) As Double	'ignore
	Dim amt As Double = pars(0)
	Return z * amt
End Sub

Private Sub minusAmt(z As Double, pars() As Object) As Double	'ignore
	Dim amt As Double = pars(0)
	Return z - amt
End Sub

Private Sub divideAmt(z As Double, pars() As Object) As Double	'ignore
	Dim amt As Double = pars(0)
	Return z / amt
End Sub

Private Sub absolute(z As Double, pars() As Object) As Double	'ignore
	Return Abs(z)
End Sub

Private Sub negative(z As Double, pars() As Object) As Double	'ignore
	Return -z
End Sub

Private Sub square(z As Double, pars() As Object) As Double	'ignore
	Return z * z
End Sub

Private Sub squareRoot(z As Double, pars() As Object) As Double	'ignore
	Return Sqrt(z)
End Sub

Private Sub genericColumn(function As String, xnr As Int) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, xnr, nc, 0)
	newM.x = CallSub(Me, function & "X")
	Return methods(newM, function & "C")
End Sub

Private Sub genericColumn1(function As String, xnr As Int, sample As Boolean) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, xnr, nc, 0)
	newM.x = CallSub2(Me, function & "X", sample)
	Return methods(newM, function & "C")
End Sub

Private Sub genericColumn2(function As String, mat As dMatrix, xnr As Int, xnc As Int, index As Int) As dMatrix
	If isEmpty Then Return Me
	If nr <> mat.nr Then 
		Return errorMessage($"${function}: The number of items in the first matrix ${id} is not comensurate with the second matrix ${mat.id}; (${nr}) versus (${mat.nr})"$)
	End If
	Dim newM As dMatrix = dValue(id, xnr, xnc, 0)
	If index < 0 Then 
		newM.x = CallSub2(Me, function & "X", mat)
	Else
		newM.x = CallSub3(Me, function & "X", mat.x, index)
	End If
	Return methods(newM, function & "C." & mat.id)
End Sub

'The following are the same for both column and row variants (row operations are applied to the transpose)
'The X version holds the algorithm. C=column  R=rows   T=total/over all elements
Private Sub maxX As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim myMax As Double = x(0, j)
		For i = 1 To nr - 1
			If x(i, j) > myMax Then myMax = x(i, j)
		Next
		newX(0, j) = myMax
	Next
	Return newX
End Sub

'for each column, find the maximum, and returns a 1xn dMatrix object
Public Sub maxC As dMatrix
	Return genericColumn("max", 1)
End Sub

'for each row, find the maximum, and returns a mx1 dMatrix object
Public Sub maxR As dMatrix
	Dim newM As dMatrix = transpose.maxC.transpose
	Return methods(newM, "maxR")
End Sub

'find the maximum of all elements and returns a 1x1 dMatrix object
Public Sub maxT As dMatrix
	If isEmpty Then Return Me
	Dim newX(,) As Double = maxX
	Dim myMax As Double = x(0, 0)
	For j = 1 To nc - 1	
		If newX(0, j) > myMax Then myMax = newX(0, j)
	Next
	Dim newM As dMatrix = dScalar(id, myMax)
	Return methods(newM, "maxT")
End Sub

Private Sub minX As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim myMin As Double = x(0, j)
		For i = 1 To nr - 1
			If x(i, j) < myMin Then myMin = x(i, j)
		Next
		newX(0, j) = myMin
	Next
	Return newX
End Sub

'for each column, find the minimum, and returns a 1xn dMatrix object
Public Sub minC As dMatrix
	Return genericColumn("min", 1)
End Sub

'for each row, find the minimum, and returns a mx1 dMatrix object
Public Sub minR As dMatrix
	Dim newM As dMatrix = transpose.minC.transpose
	Return methods(newM, "minR")
End Sub

'find the minimum of all elements and returns a 1x1 dMatrix object
Public Sub minT As dMatrix
	If isEmpty Then Return Me
	Dim newX(,) As Double = minX
	Dim myMin As Double = x(0, 0)
	For j = 1 To nc - 1	
		If newX(0, j) > myMin Then myMin = newX(0, j)
	Next
	Dim newM As dMatrix = dScalar(id & ".minT", myMin)
	Return methods(newM, "minT")
End Sub

Private Sub sumX As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim sum, num As Double
		For i = 0 To nr - 1
			If NotMissing(x(i, j)) Then 
				sum = sum + x(i, j)
				num = num + 1
			End If
		Next
		If num > 0 Then newX(0, j) = nr * sum / num
	Next
	Return newX
End Sub

'computes the sum of each column (adjusting for missing values) and returns a 1xn dMatrix object
Public Sub sumC As dMatrix
	Return genericColumn("sum", 1)
End Sub

'computes the sum of each row (adjusting for missing values) and returns a mx1 dMatrix object
Public Sub sumR As dMatrix
	Dim newM As dMatrix = transpose.sumC.transpose
	Return methods(newM, "sumR")
End Sub

'computes the sum of all elements (adkusting for missing values) and returns a 1x1 dMatrix object
Public Sub sumT As dMatrix
	If isEmpty Then Return Me
	Dim newX(,) As Double = sumX
	Dim sum, num As Double
	For j = 0 To nc - 1	
		If NotMissing(x(0, j)) Then 
			num = num + 1
			sum = sum + newX(0, j)
		End If
	Next
	Dim newM As dMatrix = dScalar(id & ".sumT", nr * sum / num)
	Return methods(newM, "sumT")
End Sub

Private Sub meanX As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim sum, num As Double
		For i = 0 To nr - 1
			If NotMissing(x(i, j)) Then 
				sum = sum + x(i, j)
				num = num + 1
			End If
		Next
		If num > 0 Then newX(0, j) = sum / num
	Next
	Return newX
End Sub

'computes the mean (average) of each column (adjusting for missing values) and returns a 1xn dMatrix object
Public Sub meanC As dMatrix
	Return genericColumn("mean", 1)
End Sub

'computes the mean (average) of each row (adjusting for missing values) and returns a mx1 dMatrix object
Public Sub meanR As dMatrix
	Dim newM As dMatrix = transpose.meanC.transpose
	Return methods(newM, "meanR")
End Sub

'computes the mean (average) over all elements (adjusting for missing values) and returns a 1x1 dMatrix object
Public Sub meanT As dMatrix
	If isEmpty Then Return Me
	Dim newX(,) As Double = sumX
	Dim sum, num, mean As Double
	For j = 0 To nc - 1	
		If NotMissing(x(0, j)) Then 
			num = num + 1
			sum = sum + newX(0, j)
		End If
	Next
	If num > 0 Then mean = sum / num
	Dim newM As dMatrix = dScalar(id & ".meanT", mean)
	Return methods(newM, "sumT")
End Sub

Private Sub varianceX(sample As Boolean) As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim sum, sumsq, num As Double
		For i = 0 To nr - 1
			Dim xf As Double = x(i, j)
			If NotMissing(xf) Then 
				num = num + 1
				sum = sum + xf
				sumsq = sumsq + xf * xf
			End If
		Next
		Dim n As Float = num
		If sample Then n = num - 1
		If n > 0 Then newX(0, j) = (sumsq - sum * sum / num) / n
	Next
	Return newX
End Sub

'computes the sample variance of each column (adjusting for missing values) and returns a 1xn dMatrix object
'if sample is False then the population variance is computed
Public Sub varianceC(sample As Boolean) As dMatrix
	Return genericColumn1("variance", 1, sample)
End Sub

'computes the sample variance of each row (adjusting for missing values) and returns a mx1 dMatrix object
'if sample is False then the population variance is computed
Public Sub varianceR(sample As Boolean) As dMatrix
	Dim newM As dMatrix = transpose.varianceC(sample).transpose
	Return methods(newM, "varianceR")
End Sub

'computes the sample variance over all elements (adjusting for missing values) and returns a 1x1 dMatrix object
'if sample is False then the population variance is computed
Public Sub varianceT(sample As Boolean) As dMatrix
	If isEmpty Then Return Me
	Dim mean As Double = meanT.x(0,0)
	Dim sum, num As Double = 0
	For i = 0 To nr - 1
		For j = 0 To nc - 1	
			If NotMissing(x(i, j)) Then 
				Dim z As Double = x(i, j) - mean
				num = num + 1
				sum = sum + z * z
			End If
		Next
	Next
	Dim newM As dMatrix 
	Dim n As Float = num
	If sample Then n = num - 1
	If n > 0 Then 
		newM = dScalar(id & ".varianceT", sum / (num * n))
	Else
		newM = dScalar(id & ".varianceT", 0)
	End If
	Return methods(newM, "varianceT")
End Sub

'computes the sample standard deviation of each column (adjusting for missing values) and returns a 1xn dMatrix object
'if sample is False then the population standard deviation is computed
Public Sub stdDevC(sample As Boolean) As dMatrix
	If isEmpty Then Return Me
	Return varianceC(sample).sqRoot
End Sub

'computes the sample standard deviation of each row (adjusting for missing values) and returns a mx1 dMatrix object
'if sample is False then the population standard deviation is computed
Public Sub stdDevR(sample As Boolean) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = varianceR(sample).sqRoot
	Return methods(newM, "stdDevR")
End Sub

'computes the sample standard deviation over all elements (adjusting for missing values) and returns a 1x1 dMatrix object
'if sample is False then the population standard deviation is computed
Public Sub stdDevT(sample As Boolean) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = varianceT(sample).sqRoot
	Return methods(newM, "stdDevT")
End Sub

Private Sub correlationX As Double(,)		'ignore
	alertCount = 0
	Dim newX(nc, nc) As Double
	Dim corr As Double
	For j1 = 0 To nc - 2
		For j2 = j1 + 1 To nc - 1
			corr = pearsonPairs(Me, j1, Me, j2)
			newX(j1, j2) = corr
			newX(j2, j1) = corr
		Next
		newX(j1, j1) = 1.0
	Next
	newX(nc - 1, nc - 1) = 1.0
	Return newX
End Sub

'computes the correlation between each pair of columns and returns the correlation matrix
'pairwise deletion of missing values
Public Sub correlationC As dMatrix
	Return genericColumn("correlation", nc)
End Sub

'computes the correlation between each pair of rows and returns the correlation matrix
'pairwise deletion of missing values
Public Sub correlationR As dMatrix
	Dim newM As dMatrix = transpose.correlationC.transpose
	Return methods(newM, "correlationR")
End Sub

Private Sub covarianceX As Double(,)		'ignore
	Dim newX(nc, nc) As Double
	Dim result() As Double
	For j1 = 0 To nc - 2
		For j2 = j1 + 1 To nc - 1
			result = sumsquared(Me, j1, Me, j2)
			Dim xnum As Double = result(3)
			newX(j1, j2) = result(0)/ (xnum - 1)
			newX(j2, j1) = newX(j1, j2)
		Next
	Next
	For j = 0 To nc - 1
		newX(j, j) = sumsquared(Me, j, Me, j)(0) / (xnum - 1)
	Next
	Return newX
End Sub

'computes the covariance between each pair of columns and returns the covariance matrix
'pairwise deletion of missing values
Public Sub covarianceC As dMatrix
	Return genericColumn("covariance", nc)
End Sub

'computes the covariance between each pair of rows and returns the covariance matrix
'pairwise deletion of missing values
Public Sub covarianceR As dMatrix
	Dim newM As dMatrix = transpose.covarianceC.transpose
	Return methods(newM, "covarianceR")
End Sub

Private Sub correlateWithX(mat As dMatrix) As Double(,)		'ignore
	Dim newX(1, mat.nc) As Double
	Dim corr As Double
	For j = 0 To mat.nc - 1
		corr = pearsonPairs(Me, 0, mat, j)
		newX(0, j) = corr
	Next
	Return newX
End Sub

'computes correlations with a single column matrix and returns the 1xn dMatrix object
'pairwise deletion of missing values
Public Sub correlateWithC(mat As dMatrix) As dMatrix
	Return genericColumn2("correlateWith", mat, 1, mat.nc, -1)
End Sub

'computes correlations with a single row matrix and returns the mx1 dMatrix object
'pairwise deletion of missing values
Public Sub correlateWithR(mat As dMatrix) As dMatrix
	Dim newM As dMatrix = transpose.correlationC.transpose
	Return methods(newM, "correlateWithR." & mat.id)
End Sub

Private Sub insertX(newCol(,) As Double, position As Int) As Double(,)		'ignore
	Dim newX(nr, nc+1) As Double
	For i = 0 To nr - 1
		For j = 0 To nc
			If j = position Then 
				newX(i, j) = newCol(i, 0)
			Else if j < position Then 
				newX(i, j) = x(i, j)
			Else 	
				newX(i, j) = x(i, j - 1)
			End If
		Next
	Next
	Return newX
End Sub

'inserts a single column dMatrix object into column index position - returns a new dMatrix object
Public Sub insertC(mat As dMatrix, index As Int) As dMatrix
	Return genericColumn2("insert", mat, nr, nc + 1, index)
End Sub

'inserts a single row dMatrix object into row index position - returns a new dMatrix object
Public Sub insertR(mat As dMatrix, index As Int) As dMatrix
	Dim newM As dMatrix = transpose.insertC(mat.transpose, index).transpose
	Return methods(newM, "insertR." & mat.id)
End Sub

Private Sub appendX(mat As dMatrix) As Double(,)		'ignore
	Dim newX(nr, nc + mat.nc) As Double
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			newX(i, j) = x(i, j)
		Next
		For j = 0 To mat.nc - 1
			newX(i, nc + j) = mat.x(i, j)
		Next
	Next
	Return newX
End Sub

'Appends a dMatrix object with the same number of rows - returns a new dMatrix object
Public Sub appendC(mat As dMatrix) As dMatrix
	Return genericColumn2("append", mat, nr, nc + mat.nc, -1)
End Sub

'Appends a dMatrix object with the same number of columns - returns a new dMatrix object
Public Sub appendR(mat As dMatrix) As dMatrix
	Dim newM As dMatrix = transpose.appendC(mat.transpose).transpose
	Return methods(newM, "appendR." & mat.id)
End Sub

'Matrix multiplication "dot" is the matrix operator (rows of the 1st matrix are multiplied by columns of the 2nd matrix and summed)
'It is named after "dot product" in vector notation - the row vectors of matrix1 are "dot product"ed with the columns of matrix2
'constraints - the #columns of the first has to be equal to the #rows of the second matrix
Public Sub dot(M As dMatrix) As dMatrix
	If isEmpty Then Return Me
	If M.nr = 0 Or M.nc = 0 Then Return errorMessage($"?${M.id} is empty."$)
	If nc <> M.nr Then 
		Return errorMessage($"dot: The # of columns (${nc}) of the first matrix ${id} is not the same as the # of rows (${M.nr}) of the second matrix"$)
	End If
	Dim newM As dMatrix = dValue(id, nr, M.nc, 0)
	For i = 0 To newM.nr - 1
		For j = 0 To newM.nc - 1
			Dim sum, num As Double
			For k = 0 To nc - 1
				If NotMissing(x(i, k)) And NotMissing(M.x(k, j)) Then 
					num = num + 1					
					sum = sum + x(i, k) * M.x(k, j)
				End If
			Next
			If num > 0 Then 
				newM.x(i, j) = nc * sum /num 		'to accomodate missing values
			Else
				newM.x(i, j) = 0/0
			End If
		Next
	Next
	Return methods(newM, "dot." & M.id)
End Sub

'Used by 'inverse' but can also be called on its own - it has useful information about a matrix
'Check Wiki: Row Echelon Form of a matrix, and Matrix Inverse
Public Sub rowEchelon As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = iterate(Me, "clone", Array())
	Dim newX(,) As Double = newM.x
	Dim h, k As Int
	Do While h < nr And k < nc
		Dim maxv As Double = 0
		Dim imax As Int = 0
		For i = h To nr - 1
			If Abs(newX(i, k)) > maxv Then 
				maxv = Abs(newX(i, k))
				imax = i
			End If
		Next
		If maxv = 0 Then 
			k = k + 1
		Else
			For j = 0 To nc - 1
				Dim temp As Double = newX(h, j)
				newX(h, j) = newX(imax, j)
				newX(imax, j) = temp
			Next
		
			For i = h + 1 To nr - 1
				Dim f As Double = newX(i, k) / newX(h, k)
				newX(i, k) = 0
				For j = k + 1 To nc - 1
					newX(i, j) = newX(i, j) - newX(h, j) * f
				Next
			Next
			h = h + 1
			k = k + 1
		End If
	Loop
	For i = nr - 1 To 0 Step -1
		Dim firstNonZero As Int =  -1
		For j = 0 To nc - 1
			If newX(i, j) <> 0 Then 
				firstNonZero = j
				Exit
			End If
		Next
		If firstNonZero > -1 Then 
			Dim f As Double = newX(i, firstNonZero)
			For j = firstNonZero To nc - 1
				newX(i, j) = newX(i, j) / f
			Next
			For ii = i - 1 To 0 Step -1
				f = newX(ii, firstNonZero)
				For j = 0 To nc - 1
					newX(ii, j) = newX(ii, j) - newX(i, j) * f
				Next
			Next
		End If
	Next
	newM.x = newX
	Return methods(newM, "rowEchelon")
End Sub

'computes the inverse as a new dMatrix object - a matrix 'dot' multiplied by it's inverse results in the identity matrix
Public Sub inverse As dMatrix
	If isEmpty Then Return Me
	If nr < nc Then 
		Dim A As dMatrix = Me
		Dim AT As dMatrix = A.transpose
		Return AT.dot(A.dot(AT).inverse)
	Else if nr > nc Then 
		Dim A As dMatrix = Me
		Dim AT As dMatrix = A.transpose
		Return AT.dot(A).inverse.dot(AT)
	End If
	Dim rows() As Int = iInx(nr, 0)
	Dim cols() As Int = iInx(nc, nc)
	Dim iz As dMatrix = dIdentity("", nc)
	Dim augM As dMatrix = appendC(iz)
	Dim newM As dMatrix = augM.rowEchelon.getSubset(rows, cols)
	Return methods(newM, "inverse")
End Sub

Private Sub pearsonPairs(M1 As dMatrix, j1 As Int, M2 As dMatrix, j2 As Int) As Double
	Dim sumSq() As Double = sumsquared(M1, j1, M2, j2)
	Dim pearsonR As Double
	If sumSq(1) = 0 And sumSq(2) = 0 Then
		errorMessage($"?${byRowCol}Correlations: matrix ${id} ${byRowCol}s ${j1} and ${j2} are Amounts, correlations with these ${byRowCol}s are invalid."$)
	else If sumSq(1) = 0 Then 
		errorMessage($"?${byRowCol}Correlations: matrix ${id} ${byRowCol} ${j1} is a Amount, correlations with this ${byRowCol} are invalid."$)
	Else if sumSq(2) = 0 Then 
		errorMessage($"?${byRowCol}Correlations: matrix ${id} ${byRowCol} ${j2} is a Amount, correlations with this ${byRowCol} are invalid."$)
	Else
		pearsonR = sumSq(0) / Sqrt(sumSq(1) * sumSq(2))
	End If
	If Abs(pearsonR) > .9 And nc > 2 And alertCount = 0 Then 
		alertCount = alertCount + 1
		errorMessage($"!${byRowCol}Correlations: Correlation between ${byRowCol}s ${j1} and ${j2} in matrix ${id} is greater than 0.9, which may affect validity of regression operations."$)
	End If
	Return pearsonR
End Sub

Private Sub sumsquared(M1 As dMatrix, j1 As Int, M2 As dMatrix, j2 As Int) As Double()
	Dim sum1, sum2, sumsq1, sumsq2, sumx1x2 As Double
	Dim x1, x2, xnum As Double
	For i = 0 To nr - 1
		x1 = M1.x(i, j1)
		x2 = M2.x(i, j2)
		If NotMissing(x1) And NotMissing(x2) Then 
			xnum = xnum + 1.0
			sum1 = sum1 + x1
			sum2 = sum2 + x2
			sumsq1 = sumsq1 + x1 * x1
			sumsq2 = sumsq2 + x2 * x2
			sumx1x2 = sumx1x2 + x1 * x2
		End If
	Next
	Return Array As Double(sumx1x2 - sum1 * sum2 / xnum , sumsq1 - sum1 * sum1 / xnum , sumsq2 - sum2 * sum2 / xnum, xnum)
End Sub

'returns a list of triplets corresponding to the location of the finds - i, j, value
'multiple values can be search for and multiple hits can occur for each
Public Sub find(values() As Double) As List
	Dim a As List: a.Initialize
	For Each z As Double In values
		For i = 0 To nr - 1
			For j = 0 To nc - 1
				If x(i, j) = z Or (x(i, j) <> x(i, j) And z <> z) Then  
					Dim res(3) As Double
					res(0) = i: res(1) = j: res(2) = z
					a.Add(res)
				End If
			Next
		Next
	Next
	Return a
End Sub

'returns an integer representing the number of linear independent rows/columns
Public Sub rank As Int
	Dim newM As dMatrix = rowEchelon
	Dim rnk As Int = 0
	For i = 0 To nr - 1
		Dim zero As Boolean = True
		For j = 0 To nc - 1
			If newM.x(i, j) <> 0 Then 
				zero = False
				Exit
			End If
		Next
		If zero = False Then rnk = rnk + 1
	Next
	Return rnk
End Sub

'returns True if all elements of the matrix are within a small amount (defined by eps) of the other
Public Sub withinEps(mat As dMatrix, eps As Double) As Boolean
	If nr<>mat.nr Or nc<>mat.nc Then Return False
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			If Abs(x(i, j) - mat.x(i,j)) > eps Then Return False
		Next
	Next
	Return True
End Sub

'returns a sorted dMatrix object, sorted by column, either in ascending or descending order
Public Sub sortC(index As Int, ascending As Boolean) As dMatrix
	If isEmpty Then Return Me
	Dim newM As dMatrix = dValue(id, nr, nc, 0)
	Dim sortList As List: sortList.Initialize
	For i = 0 To newM.nr - 1
		Dim spair As sortPairDouble
		spair.initialize
		spair.index = i
		spair.value = x(i, index) 
		sortList.add(spair)
	Next
	sortList.SortType("value", ascending)
	For i = 0 To nr - 1
		Dim spair As sortPairDouble = sortList.Get(i)
		For j = 0 To nc - 1
			newM.x(i, j) = x(spair.index, j)
		Next
	Next
	Return methods(newM, "sortC." & index)
End Sub

'returns a sorted dMatrix object, sorted by row, either in ascending or descending order
Public Sub sortR(index As Int, ascending As Boolean) As dMatrix
	Dim newM As dMatrix = transpose.sortC(index, ascending).transpose
	Return methods(newM, "sortR." & index)
End Sub

'converts a dMAtrix object to a vector of values column-wise
Public Sub toVectorC As Double()
	Dim res(nr*nc) As Double
	Dim index As Int = -1
	For j = 0 To nc - 1
		For i = 0 To nr - 1
			index = index + 1
			res(index) = x(i, j)
		Next
	Next
	Return res
End Sub

'converts a dMAtrix object to a vector of values row-wise
Public Sub toVectorR As Double()
	Dim res(nr*nc) As Double
	Dim index As Int = -1
	For i = 0 To nr - 1
		For j = 0 To nc - 1
			index = index + 1
			res(index) = x(i, j)
		Next
	Next
	Return res
End Sub

Private Sub NotMissing(z As Double) As Boolean
	Return z = z
End Sub

Private Sub validX As Double(,)		'ignore
	Dim newX(1, nc) As Double
	For j = 0 To nc - 1
		Dim num As Double = 0
		For i = 0 To nr - 1
			If NotMissing(x(i, j)) Then num = num + 1
		Next
		newX(0, j) = num
	Next
	Return newX
End Sub

'computes the number of valid (non-missing) elements in each column and returns a 1xn dMatrix object 
Public Sub validC As dMatrix
	Return genericColumn("valid", 1)
End Sub

'computes the number of valid (non-missing) elements in each row and returns a mx1 dMatrix object 
Public Sub validR As dMatrix
	Dim newM As dMatrix = transpose.validC.transpose
	Return methods(newM, "validR")
End Sub

Private Sub validPairsX As Double(,)		'ignore
	alertCount = 0
	Dim newX(nc, nc) As Double
	Dim xnum As Int
	For j1 = 0 To nc - 1
		For j2 = j1 To nc - 1
			xnum = 0
			For i = 0 To nr - 1
				If NotMissing(x(i, j1)) And NotMissing(x(i, j2)) Then xnum = xnum + 1
			Next
			newX(j1, j2) = xnum
			newX(j2, j1) = xnum
		Next
	Next
	Return newX
End Sub

'computes the number of valid (non-missing) pair in each pair of columns and returns a dMatrix object 
Public Sub validPairsC As dMatrix
	Return genericColumn("validPairs", nc)
End Sub

'computes the number of valid (non-missing) pair in each pair of rows and returns a dMatrix object 
Public Sub validPairsR As dMatrix
	Dim newM As dMatrix = transpose.validPairsC.transpose
	Return methods(newM, "validPairsR")
End Sub

Private Sub compareGT(z As Double, pars() As Object) As Double	'ignore
	Dim amt As Double = pars(0)
	If z > amt Then Return 1 Else Return 0
End Sub

Private Sub toInx(i As Int) As Int()	'ignore
	Return Array As Int(i)
End Sub

'returns an array of sequential double precision numbers, starting with offset
Public Sub dInx(n As Int, offset As Double) As Double()
	Dim res(n) As Double
	For i = 0 To n - 1
		res(i) = offset + i 		
	Next
	Return res
End Sub

'returns an array of sequential integers, starting with startAt
Public Sub iInx(n As Int, startAt As Int) As Int()
	Dim resi(n) As Int
	For i = 0 To n - 1
		resi(i) = startAt + i 		
	Next
	Return resi
End Sub

Private Sub reflect(z As Double, pars() As Object) As Double	'ignore
	Dim comp As Double = pars(0)
	If comp < 0 Then 
		Return -z
	Else
		Return z	
	End If
End Sub

'converts a matrix to a List, each item in the list is a row vector
Public Sub asList(mat(,) As Double) As List
    Dim Jo As JavaObject
    Jo.InitializeStatic("java.util.Arrays")
	Return Jo.RunMethod("asList",Array(mat))
End Sub
